/* Copyright (c) 2007 Michele Bini
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA */


#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>

/* (2**255) - 19 =
   0x7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffed

   0111 + (1111) x 7 + (1111) x (8*6) + (1111) x 6 + 1110 + 1101
   0x7fffffff ffffffff ffffffff ffffffff ffffffff ffffffff 0xffffffff 0xffffffed
*/

/* todo: conditional compilation when mp_limb_t > 32 bits */
#define P25519BITS 256
#define P25519USEDBITS (P25519BITS - 1)
#define P25519N (P25519BITS/32)
typedef mp_limb_t cmodp_t[P25519N];
static cmodp_t p25519 = { 0xffffffed, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0x7fffffff };
static cmodp_t zerocmp = { 0, 0, 0, 0, 0, 0, 0, 0 };
static cmodp_t onecmp = { 1, 0, 0, 0, 0, 0, 0, 0 };

static int
getbit(cmodp_t *x, unsigned int n) {
  unsigned int d = (sizeof(zerocmp[0])*8);
  return (x[0][n / d] >> (n % d)) & 1;
}

static void
setbit(cmodp_t *x, unsigned int n, int v) {
  unsigned int d = (sizeof(zerocmp[0])*8);
  unsigned int i = n / d;
  mp_limb_t l = x[0][i];
  if (v) {
    l |= 1<<(n % d);
  } else {
    l &= ~((mp_limb_t)(1<<(n % d)));
  }
  x[0][i] = l;
}

char base32chars[32] = {
  'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h',
  'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p',
  'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
  'y', 'z', '2', '3', '4', '5', '6', '7'
};

static void
base32_encode(char*s, cmodp_t *x) {
  int c; int i = 0; int v = 0; int a = 0;
  for (c = 0; c < P25519BITS; c++) {
    v |= getbit(x, c) << i;
    /* fprintf(stderr, "v: %x\n", v); */
    i++;
    if (i > 4) {
      s[a] = base32chars[v];
      v = 0;
      i = 0;
      a++;
    }
  }
  if (v) {
    s[a] = base32chars[v];
    a++;
  }
  s[a] = 0;
  a--;
  i=0;
  while (i < a) {
    v = s[a];
    s[a] = s[i];
    s[i] = v;
    i++; a--;
  }
}

static void
base32_decode(const char*s, cmodp_t *x) {
  int c; int i = 0; int v = 0; int a = 0;
  while (s[a]) {
    a++;
  }
  a--;
  while (a >= 0) {
    c = s[a];
    switch (c) {
    case '2': case '3': case '4': case '5': case '6': case '7':
      v = (c - '2')+26;
      break;
    default:
      v = c - 'a';
    }
    setbit(x, i, v&1); v>>=1; i++; if (i >= P25519BITS) { break; };
    setbit(x, i, v&1); v>>=1; i++; if (i >= P25519BITS) { break; };
    setbit(x, i, v&1); v>>=1; i++; if (i >= P25519BITS) { break; };
    setbit(x, i, v&1); v>>=1; i++; if (i >= P25519BITS) { break; };
    setbit(x, i, v&1); v>>=1; i++; if (i >= P25519BITS) { break; };
    a--;
  }
  while (i < P25519BITS) {
    setbit(x, i, 0);
    i++;
  }
}

#if 0
static void
tracev(char*m, cmodp_t *x) {
  char s[(P25519BITS/4)+2];
  base32_encode(s, x);
  fprintf(stderr, "%s: %s\n", m, s);
}
#endif

static void
copycmp(cmodp_t *n, cmodp_t *x) {
  int c;
  for (c = 0; c < P25519N; c++) {
    n[0][c] = x[0][c];
  }
}

static
int zeromodp(cmodp_t *x) {
  return (mpn_cmp((mp_limb_t*)x, (mp_limb_t*)&zerocmp, P25519N) == 0);
}

static
void addmodp(cmodp_t *a, cmodp_t *b) {
  mpn_add_n((mp_limb_t*)a, (mp_limb_t*)a, (mp_limb_t*)b, P25519N);
  if (mpn_cmp((mp_limb_t*)a, (mp_limb_t*)&p25519, P25519N) >= 0) {
    mpn_sub_n((mp_limb_t*)a, (mp_limb_t*)a, (mp_limb_t*)&p25519, P25519N);
  }
}

static
void submodp(cmodp_t *a, cmodp_t *b) {
  if (mpn_cmp((mp_limb_t*)b, (mp_limb_t*)a, P25519N) > 0) {
    mpn_add_n((mp_limb_t*)a, (mp_limb_t*)&p25519, (mp_limb_t*)a, P25519N);
  }
  mpn_sub_n((mp_limb_t*)a, (mp_limb_t*)a, (mp_limb_t*)b, P25519N);
}

static void
mulmodp(cmodp_t *a, cmodp_t *b) {
  mp_limb_t d[P25519N*2]; mp_limb_t r[P25519N+1];
  mpn_mul(d, (mp_limb_t*)a, P25519N, (mp_limb_t*)b, P25519N);
  mpn_tdiv_qr(r, (mp_limb_t*)a, 0, d, P25519N*2, (mp_limb_t*)&p25519, P25519N);
}

static void
sqrmodp(cmodp_t *a) {
  mulmodp(a, a);
}

static void
invmodp(cmodp_t *a) {
  /* a = a ** (p-2)
     0111 + (1111) x 7 + (1111) x (8*6) + (1111) x 6 + 1110 + 1011
     0 . 1 x (3 + 4*7 + 4 * 8*6 + 4*6 + 3) . 0 . 1011
     0 . 1 x (250) . 0 . 1011
  */
  cmodp_t c; copycmp(&c, a);
  int i = 250;
  while (--i) {
    sqrmodp(a); mulmodp(a, &c);
  }
  sqrmodp(a);
  sqrmodp(a); mulmodp(a, &c);
  sqrmodp(a);
  sqrmodp(a); mulmodp(a, &c);
  sqrmodp(a); mulmodp(a, &c);
}

static mp_limb_t asmall = 121665; /* (486662 - 2) / 4; */

static
void mulasmall(cmodp_t *a) {
  mp_limb_t d[P25519N+1]; mp_limb_t r[2];
  //tracev("mulasmall a", a);
  d[P25519N] = mpn_mul_1(d, (mp_limb_t*)a, P25519N, asmall);
  //tracev("mulasmall d", d);
  //tracev("mulasmall a", a);
  mpn_tdiv_qr(r, (mp_limb_t*)a, 0, d, P25519N+1, (mp_limb_t*)&p25519, P25519N);
  //tracev("mulasmall a", a);
}

static
void dbl(cmodp_t *x2, cmodp_t *z2, cmodp_t *x, cmodp_t *z) {
  cmodp_t m, n, o;
  //tracev("dbl x", x);
  //tracev("dbl z", z);
  copycmp(&m, x); addmodp(&m, z); sqrmodp(&m);
  //tracev("dbl m", &m);
  copycmp(&n, x); submodp(&n, z); sqrmodp(&n);
  //tracev("dbl n", &n);
  copycmp(&o, &m); submodp(&o, &n);
  //tracev("dbl o", &o);
  copycmp(x2, &n); mulmodp(x2, &m);
  copycmp(z2, &o); mulasmall(z2); addmodp(z2, &m); mulmodp(z2, &o);
  //tracev("dbl x2", x2);
  //tracev("dbl z2", z2);
}

static
void sum(cmodp_t *x3, cmodp_t *z3, cmodp_t *x, cmodp_t *z, cmodp_t *xp, cmodp_t *zp, cmodp_t *x1) {
  cmodp_t k, l, p, q;
  //tracev("sum x", x);
  //tracev("sum z", z);
  copycmp(&p, x); submodp(&p, z); copycmp(&k, xp); addmodp(&k, zp); mulmodp(&p, &k);
  copycmp(&q, x); addmodp(&q, z); copycmp(&l, xp); submodp(&l, zp); mulmodp(&q, &l);
  //tracev("sum p", &p);
  //tracev("sum q", &q);
  copycmp(x3, &p); addmodp(x3, &q); sqrmodp(x3);
  copycmp(z3, &p); submodp(z3, &q); sqrmodp(z3); mulmodp(z3, x1);
}

extern void
curve25519(cmodp_t *r, cmodp_t *f, cmodp_t *c) {
  cmodp_t x1, xa, za, x, z;
  //tracev("f", f);
  if (zeromodp(f)) {
    copycmp(r, &zerocmp);
    return;
  }
  copycmp(&x1, c);
  //tracev("c", c);
  //tracev("x1", x1);
  dbl(&xa, &za, &x1, &onecmp);
  //tracev("xa", &xa);
  //tracev("za", &za);
  copycmp(&x, &x1);
  copycmp(&z, &onecmp);

  int n = P25519BITS-1;

  while (getbit(f, n) == 0) {
    n--;
  }
  n--;

  while (n >= 0) {
    cmodp_t nx, nz, nxa, nza;
    int b = getbit(f, n);
    //fprintf(stderr, "b: %d\n", b);
    if (b == 0) {
      dbl(&nx, &nz, &x, &z);
      sum(&nxa, &nza, &xa, &za, &x, &z, &x1);
    } else {
      sum(&nx, &nz, &xa, &za, &x, &z, &x1);
      dbl(&nxa, &nza, &xa, &za);
    }
    copycmp(&x, &nx); copycmp(&z, &nz); copycmp(&xa, &nxa); copycmp(&za, &nza);
    //tracev("xn", &x);
    //tracev("zn", &z);
    //tracev("xa", &xa);
    //tracev("za", &za);
    n--;
  }

  //tracev("x", &x);
  //tracev("z", &z);
  invmodp(&z);
  //tracev("1/z", &z);
  mulmodp(&x, &z);
  copycmp(r, &x);
}

int main(int argc, const char *argv[]) {
  char s[(P25519BITS/4)+2];
  cmodp_t c, f, r;
  if (argc != 3) {
    fprintf(stderr, "Usage: %s base32encodedprivatekey base32encodedpublickey\n", argv[0]);
    exit(EXIT_FAILURE);
  }
  base32_decode(argv[1], &f);
  base32_decode(argv[2], &c);
  curve25519(&r, &f, &c);
  //tracev("f", &f);
  //tracev("c", &c);
  base32_encode(s, &r); printf("%s\n", s);
#if 0
  char s[(P25519BITS/4)+2];
  cmodp_t v;
  base32_encode(s, &p25519);
  printf("%s\n", s);
  base32_decode(s, &v);
  base32_encode(s, &v);
  printf("%s\n", s);
#endif
  exit(EXIT_SUCCESS);
}
